(* Lazy Value *)

module Delayed : sig
  type 'a delayed
  exception Lazy_Error
  val delay : (unit -> 'a) -> 'a delayed
  val force : 'a delayed -> 'a
end = struct
  type 'a delayed = (unit -> 'a)
  exception Lazy_Error
      
  let delay f = f
      
  let force l = l ()
end

(* Lazy Sequence *)

module type SEQUENCE = 
  sig
    exception Empty
    type 'a seq (* abstract datatype for sequence *)
	  
    val head : 'a seq -> 'a
      (* return the head element of the sequence *)
	
    val tail : 'a seq -> 'a seq
      (* discard the head element from the sequence *)
	
    val nil : 'a seq
      (* empty sequence *)
	
    val is_empty : 'a seq -> bool
      (* return true if the sequence is empty *)
	
    val cons : 'a -> 'a seq Delayed.delayed -> 'a seq
      (* construct the sequence from the first element *)
      (* and the generator of the rest elements *)
	
    val take : 'a seq -> int -> 'a list
      (* return first n elements as list. *)
  end      

