open LazyMiniML

(* let failwith s = raise (Failure s) *)

let rec get id = function
    [] -> raise (Failure ("Unbound value " ^ id))
  | (i, v) :: ls ->
      if i = id then 
	(match !v with
	  {contents=Value Undefined} -> failwith "use undefined value with letrec"
	| x -> x)
      else get id ls

let get_int = function
    Int(v1) -> v1 | _ -> failwith "not an int value"

let get_bool = function
    Bool(v1) -> v1 | _ -> failwith "not a bool value"

exception MatchFailed

(* val eval : env * expr -> mlvalue *)

let make_delayed env e = (ref (Lazy(e, env)) : lazymlvalue)
let make_value v = (ref (Value(v)) : lazymlvalue)

let delay env e = make_delayed env e
let update ((e: lazymlvalue), e1) = e := e1

let rec eval env e = force (make_delayed env e)
and force ed = 
  match !ed with
    Value v -> v
  | Lazy (e, env) ->
    update (ed, (force1 (env, e))); force ed
and force1 (env, e) = match e with
    Const(v) -> Value v
  | Var(id) -> Value(force(get id env))
  | Plus(e1, e2) -> 
      let e1', e2' = eval env e1, eval env e2 in
      Printf.printf "(%d + %d)\n" (get_int e1') (get_int e2');
      Value(Int(get_int e1' + get_int e2'))
  | Minus(e1, e2) -> 
      let e1', e2' = eval env e1, eval env e2 in
      Printf.printf "(%d - %d)\n" (get_int e1') (get_int e2');
      Value(Int(get_int e1' - get_int e2'))
  | Times(e1, e2) -> 
      let e1', e2' = eval env e1, eval env e2 in
      Printf.printf "(%d * %d)\n" (get_int e1') (get_int e2');
      Value(Int(get_int e1' * get_int e2'))
  | Div(e1, e2) -> 
      let e1', e2' = eval env e1, eval env e2 in
      Printf.printf "(%d / %d)\n" (get_int e1') (get_int e2');
      Value(Int(get_int e1' / get_int e2'))
  | Equal(e1, e2) ->
      let e1', e2' = eval env e1, eval env e2 in
      Value(Bool(e1' = e2'))
  | ConsExp(e1, e2) ->
      Value(Cons(delay env e1, delay env e2))
  | PairExp(e1, e2) ->
      Value(Pair(delay env e1, delay env e2))
  | IfExp(ec, e1, e2) ->
      ...
  | LambdaExp [p1, e] -> 
      Value(Closure([p1, e], env))
  | LambdaExp ps -> 
      Value(Closure(ps, env))
  | App(e1, e2) -> ...
  | MatchExp(exp,pats) ->
      let v1 = delay env exp in
      let rec iter = function
	  [] -> failwith "pattern match failed at match expression"
	| (pat, body) :: rest ->
	    try
	      match_pattern pat v1, body
	    with 
	      MatchFailed -> iter rest
      in
      let binds, body = iter pats
      in Lazy(body, binds @ env)
  | LetExp([IdentPtn id, e1], e2) ->
      let v1 = delay env e1 in
      Lazy(e2, (id, ref v1) :: env)
  | LetExp(ptns, e2) ->
      let rec iter = function
	  [] -> []
	| (pat, exp) :: more_pats -> 
	    (match_pattern pat (delay env exp)) @ (iter more_pats)
      in
      Lazy(e2, (iter ptns) @ env)
  | LetRecExp([IdentPtn id, e1], e2) ->
      ...
and match_pattern pat v = 
  ...

